@@ -4,16 +4,16 @@ import math
 from itertools import combinations
 import pandas as pd
 from hierarch.internal_functions import (
-    nb_data_grabber,
     GroupbyMean,
     _compute_interval,
     bivar_central_moment,
 )
 from hierarch.resampling import Bootstrapper, Permuter
 from warnings import warn, simplefilter
+from functools import lru_cache
 
 
-def preprocess_data(data):
+def _preprocess_data(data):
     """Performs label encoding without overwriting numerical variables.
 
     Parameters
@@ -129,7 +129,7 @@ def studentized_covariance(x, y):
 
 
 @jit(nopython=True, cache=True)
-def welch_statistic(data, col: int, treatment_labels):
+def welch_statistic(sample_a, sample_b):
     """Calculates Welch's t statistic.
 
     Takes a 2D data matrix, a column to classify data by, and the labels
@@ -153,27 +153,17 @@ def welch_statistic(data, col: int, treatment_labels):
     Examples
     --------
 
-    >>> x = np.array([[0, 0, 0, 0, 0, 1, 1, 1, 1, 1], 
-    ...               [1, 2, 3, 4, 5, 10, 11, 12, 13, 14]])
-    >>> x.T
-    array([[ 0,  1],
-           [ 0,  2],
-           [ 0,  3],
-           [ 0,  4],
-           [ 0,  5],
-           [ 1, 10],
-           [ 1, 11],
-           [ 1, 12],
-           [ 1, 13],
-           [ 1, 14]])
-    >>> welch_statistic(x.T, 0, (0, 1))
+    >>> import scipy.stats as stats
+    >>> a = np.array([1, 2, 3, 4, 5])
+    >>> b = np.array([10, 11, 12, 13, 14])
+    >>> welch_statistic(a, b)
     -9.0
 
     This uses the same calculation as scipy's ttest function.
 
     >>> import scipy.stats as stats
-    >>> a = [1, 2, 3, 4, 5]
-    >>> b = [10, 11, 12, 13, 14]
+    >>> a = np.array([1, 2, 3, 4, 5])
+    >>> b = np.array([10, 11, 12, 13, 14])
     >>> stats.ttest_ind(a, b, equal_var=False)[0]
     -9.0
     
@@ -186,8 +176,7 @@ def welch_statistic(data, col: int, treatment_labels):
     https://doi.org/10.1016/S0167-7152(97)00043-6.
 
     """
-    # get our two samples from the data matrix
-    sample_a, sample_b = nb_data_grabber(data, col, treatment_labels)
+
     len_a, len_b = len(sample_a), len(sample_b)
 
     # mean difference
@@ -203,16 +192,55 @@ def welch_statistic(data, col: int, treatment_labels):
     return t
 
 
-TEST_STATISTICS = {
-    "means": welch_statistic,
-    "corr": studentized_covariance,
-}
+@lru_cache()
+def _test_stat_factory(treatment_col, compare):
+    """Prepares test statistic functions for use in hypothesis_test.
+
+    Parameters
+    ----------
+    treatment_col : 1D tuple
+        Treatment column in the design matrix. Needs to be a tuple
+        so lru_cache can work.
+    compare : {'means', 'corr'}
+        Specifies test statistic to return. 
+
+    Returns
+    -------
+    function
+        Functions that come out of _test_stat_factory take the treatment
+        column of a design matrix and the dependent variable column to compute
+        a test statistic.
+
+    """
+    if compare == "means":
+        treatment_labels = np.unique(treatment_col)
+        if treatment_labels.size != 2:
+            raise ValueError("Needs 2 samples.")
+
+        @jit(nopython=True)
+        def _welch_stat(X, y):
+            sample_a, sample_b = _grabber(X, y, treatment_labels)
+            return welch_statistic(sample_a, sample_b)
+
+        return _welch_stat
+
+    if compare == "corr":
+        return studentized_covariance
 
 
-def linear_regression_test(
+@jit(nopython=True)
+def _grabber(X, y, treatment_labels):
+    slicer = X == treatment_labels[0]
+    sample_a = y[slicer]
+    sample_b = y[~slicer]
+    return sample_a, sample_b
+
+
+def hypothesis_test(
     data_array,
     treatment_col: int,
     compare="corr",
+    alternative="two-sided",
     skip=None,
     bootstraps=100,
     permutations=1000,
@@ -220,8 +248,8 @@ def linear_regression_test(
     return_null=False,
     random_state=None,
 ):
-    """Two-tailed hierarchical permutation test for any number of samples
-    with a hypothesized linear relationship.
+    """Two-tailed hierarchical permutation test for change in location
+    with any number of samples.
 
     Equivalent to calculating a p-value for a slope coefficient in a linear model.
 
@@ -237,6 +265,8 @@ def linear_regression_test(
     compare : str, optional
         The test statistic to use to perform the hypothesis test, by default "corr"
         which automatically calls the studentized covariance test statistic.
+    alternative : {"two-sided", "less", "greater"}
+        The alternative hypothesis for the test, "two-sided" by default.
     skip : list of ints, optional
         Columns to skip in the bootstrap. Skip columns that were sampled
         without replacement from the prior column, by default None
@@ -284,18 +314,20 @@ def linear_regression_test(
     >>> print(data.shape)
     (24, 4)
 
-    >>> linear_regression_test(data, treatment_col=0,
+    >>> hypothesis_test(data, treatment_col=0,
     ...                 bootstraps=1000, permutations='all',
     ...                 random_state=1)
-    0.013714285714285714
+    0.013685714285714285
 
-    This test should give the same or a very similar p-value to two_sample_test 
-    for datasets with two treatment groups.
+    By setting compare to "means", this function will perform a permutation t-test.
+    "corr", which is based on a studentized covariance test statistic, should give the 
+    same or a very similar p-value to the permutation t-test for datasets with two 
+    treatment groups.
 
-    >>> two_sample_test(data, treatment_col=0,
+    >>> hypothesis_test(data, treatment_col=0, compare='means',
     ...                 bootstraps=1000, permutations='all',
     ...                 random_state=1)
-    0.013714285714285714
+    0.013685714285714285
 
     This test can handle data with multiple treatment groups that have a 
     hypothesized linear relationship.
@@ -310,15 +342,17 @@ def linear_regression_test(
 
     There are 2,520 possible permutations, so choose a subset. 
 
-    >>> linear_regression_test(data, treatment_col=0,
+    >>> hypothesis_test(data, treatment_col=0,
     ...                 bootstraps=100, permutations=1000,
     ...                 random_state=1)
-    0.00675
+    0.00668
 
 
     """
+
+    # turns the input array or dataframe into a float64 array
     if isinstance(data_array, (np.ndarray, pd.DataFrame)):
-        data = preprocess_data(data_array)
+        data = _preprocess_data(data_array)
     else:
         raise TypeError("Input data must be ndarray or DataFrame.")
 
@@ -350,15 +384,10 @@ def linear_regression_test(
 
     # fetch test statistic from dictionary or, if given a custom test statistic, make sure it is callable
     if isinstance(compare, str):
-        try:
-            teststat = TEST_STATISTICS[compare]
-        except KeyError:
-            print(
-                "Invalid comparison. Available comparisons are: "
-                + "".join(stat for stat in TEST_STATISTICS.keys())
-            )
-            raise
-    elif not callable(compare):
+        teststat = _test_stat_factory(tuple(data[:, treatment_col].tolist()), compare)
+    elif callable(compare):
+        teststat = compare
+    else:
         raise AttributeError("Custom test statistics must be callable.")
 
     # aggregate our data up to the treated level and determine the
@@ -431,266 +460,21 @@ def linear_regression_test(
                 teststat(permute_resample[:, treatment_col], permute_resample[:, -1])
             )
 
-    # two tailed test, so check where absolute values of the null distribution
-    # are greater or equal to the absolute value of the observed difference
-    pval = np.where((np.array(np.abs(null_distribution)) >= np.abs(truediff)))[
-        0
-    ].size / len(null_distribution)
+    # generate both one-tailed p-values, then two-tailed
+    p_less = np.where(truediff > np.array(null_distribution))[0].size / len(
+        null_distribution
+    )
+    p_greater = np.where(truediff < np.array(null_distribution))[0].size / len(
+        null_distribution
+    )
+    p_two = 2 * np.min((p_less, p_greater))
 
-    if pval == 0:
-        pval += 1 / (total)
-
-    if return_null is True:
-        return pval, null_distribution
-
-    else:
-        return pval
-
-
-def two_sample_test(
-    data_array,
-    treatment_col: int,
-    compare="means",
-    skip=None,
-    bootstraps=100,
-    permutations=1000,
-    kind="weights",
-    return_null=False,
-    random_state=None,
-):
-    """Two-tailed two-sample hierarchical permutation test.
-
-    Parameters
-    ----------
-    data_array : 2D numpy array or pandas DataFrame
-        Array-like containing both the independent and dependent variables to
-        be analyzed. It's assumed that the final (rightmost) column
-        contains the dependent variable values.
-    treatment_col : int
-        The index number of the column containing "two samples" to be compared.
-        Indexing starts at 0.
-    compare : str, optional
-        The test statistic to use to perform the hypothesis test. "means" automatically
-        calls the Welch t-statistic for a difference of means test, by default "means"
-    skip : list of ints, optional
-        Columns to skip in the bootstrap. Skip columns that were sampled
-        without replacement from the prior column, by default None
-    bootstraps : int, optional
-        Number of bootstraps to perform, by default 100. Can be set to 1 for a
-        permutation test without any bootstrapping.
-    permutations : int or "all", optional
-        Number of permutations to perform PER bootstrap sample. "all"
-        for exact test, by default 1000
-    kind : str, optional
-        Bootstrap algorithm - see Bootstrapper class, by default "weights"
-    return_null : bool, optional
-        Return the null distribution as well as the p value, by default False
-    seed : int or numpy random Generator, optional
-        Seedable for reproducibility., by default None
-
-    Returns
-    -------
-    float64
-        p-value for the hypothesis test
-
-    list
-        Empirical null distribution used to calculate the p-value
-
-    Raises
-    ------
-    TypeError
-        Raised if input data is not ndarray or DataFrame.
-    ValueError
-        Raised if treatment_col has more than two different labels in it.
-    KeyError
-        If comparison is a string, it must be in the TEST_STATISTICS dictionary.
-    AttributeError
-        If comparison is a custom statistic, it must be a function.
-
-    Examples
-    --------
-    Specify the parameters of a dataset with a difference of means of 2.
-
-    >>> from hierarch.power import DataSimulator
-    >>> import scipy.stats as stats
-    >>> paramlist = [[0, 2], [stats.norm], [stats.norm]]
-    >>> hierarchy = [2, 4, 3]
-    >>> datagen = DataSimulator(paramlist, random_state=123)
-    >>> datagen.fit(hierarchy)
-    >>> data = datagen.generate()
-    >>> print(data.shape)
-    (24, 4)
-
-    >>> two_sample_test(data, treatment_col=0,
-    ...                 bootstraps=1000, permutations='all',
-    ...                 random_state=1)
-    0.03402857142857143
-
-    Instead of an exact test, a number of random permutations can be specified.
-    In this case there are 70 possible permutations.
-    
-    >>> two_sample_test(data, treatment_col=0,
-    ...                 bootstraps=1000, permutations=70,
-    ...                 random_state=1)
-    0.03357142857142857
-
-    The treatment column does not have to be the outermost column.
-
-    >>> paramlist = [[stats.norm], [0, 1]*3, [stats.norm], [stats.norm]]
-    >>> hierarchy = [3, 2, 4, 3]
-    >>> datagen = DataSimulator(paramlist, random_state=123)
-    >>> datagen.fit(hierarchy)
-    >>> data = datagen.generate()
-    >>> print(data.shape)
-    (72, 5)
-
-    Because of the larger number of possible permutations, it is usually better
-    to reduce the number of bootstraps and increase the number of permutations.
-
-    >>> two_sample_test(data, treatment_col=0,
-    ...                 bootstraps=100, permutations=1000,
-    ...                 random_state=1)
-    Traceback (most recent call last):
-        ...
-    ValueError: Needs 2 samples.
-
-    Make sure that treatment_col is set to right column index.
-
-    >>> two_sample_test(data, treatment_col=1,
-    ...                 bootstraps=100, permutations=1000,
-    ...                 random_state=1)
-    0.00276
-    """
-
-    # turns the input array or dataframe into a float64 array
-    if isinstance(data_array, (np.ndarray, pd.DataFrame)):
-        data = preprocess_data(data_array)
-    else:
-        raise TypeError("Input data must be ndarray or DataFrame.")
-
-    # set random state
-    rng = np.random.default_rng(random_state)
-
-    # enforce lower bound on skip
-    if skip is not None:
-        skip = list(skip)
-        for v in reversed(skip):
-            if v <= treatment_col + 1:
-                warn("No need to include columns before treated columns in skip.")
-                skip.remove(v)
-    else:
-        skip = []
-
-    # enforce bounds on bootstraps and permutations
-    if not isinstance(bootstraps, int) or bootstraps < 1:
-        raise TypeError("bootstraps must be an integer greater than 0")
-    if isinstance(permutations, str):
-        if permutations != "all":
-            raise TypeError("permutations must be 'all' or an integer greater than 0")
-    elif not isinstance(permutations, int) or permutations < 1:
-        raise TypeError("permutations must be 'all' or an integer greater than 0")
-
-    # initialize and fit the bootstrapper to the data
-    bootstrapper = Bootstrapper(random_state=rng, kind=kind)
-    bootstrapper.fit(data, skip=skip)
-
-    # gather labels and raise an exception if there are more than two
-    try:
-        treatment_labels = np.unique(data[:, treatment_col])
-    except IndexError:
-        print("treatment_col must be an integer")
-        raise
-    if treatment_labels.size != 2:
-        raise ValueError("Needs 2 samples.")
-
-    # fetch test statistic from dictionary or, if given a custom test statistic, make sure it is callable
-    if isinstance(compare, str):
-        try:
-            teststat = TEST_STATISTICS[compare]
-        except KeyError:
-            print(
-                "Invalid comparison. Available comparisons are: "
-                + "".join(stat for stat in TEST_STATISTICS.keys())
-            )
-            raise
-    elif not callable(compare):
-        raise AttributeError("Custom test statistics must be callable.")
-
-    # aggregate our data up to the treated level and determine the
-    # observed test statistic
-    aggregator = GroupbyMean()
-    aggregator.fit(data)
-
-    # determine the number of groupby reductions need to be done
-    levels_to_agg = data.shape[1] - treatment_col - 3
-
-    # if levels_to_agg = 0, there are no bootstrap samples to
-    # generate.
-    if (levels_to_agg - len(skip)) == 0 and bootstraps > 1:
-        bootstraps = 1
-        simplefilter("always", UserWarning)
-        warn("No levels to bootstrap. Setting bootstraps to zero.")
-
-    test = data
-    test = aggregator.transform(test, iterations=levels_to_agg)
-    truediff = teststat(test, treatment_col, treatment_labels)
-
-    # initialize and fit the permuter to the aggregated data
-    # don't need to seed this, as numba's PRNG state is shared
-    permuter = Permuter()
-
-    if permutations == "all":
-        permuter.fit(test, treatment_col, exact=True)
-
-        # in the exact case, determine and set the total number of
-        # possible permutations
-        counts = np.unique(test[:, 0], return_counts=True)[1]
-        permutations = _binomial(counts.sum(), counts[0])
-
-    else:
-        # just fit the permuter if this is a randomized test
-        permuter.fit(test, treatment_col)
-
-    # skip the dot on the permute function
-    call_permute = permuter.transform
-
-    # initialize empty null distribution list
-    null_distribution = []
-    total = bootstraps * permutations
-
-    # first set of permutations is on the original data
-    # this helps to prevent getting a p-value of 0
-    for k in range(permutations):
-        permute_resample = call_permute(test)
-        null_distribution.append(
-            teststat(permute_resample, treatment_col, treatment_labels)
-        )
-
-    # already did one set of permutations
-    bootstraps -= 1
-
-    for j in range(bootstraps):
-        # generate a bootstrapped sample and aggregate it up to the
-        # treated level
-        bootstrapped_sample = bootstrapper.transform(data, start=treatment_col + 2)
-        bootstrapped_sample = aggregator.transform(
-            bootstrapped_sample, iterations=levels_to_agg
-        )
-
-        # generate permuted samples, calculate test statistic,
-        # append to null distribution
-
-        for k in range(permutations):
-            permute_resample = call_permute(bootstrapped_sample)
-            null_distribution.append(
-                teststat(permute_resample, treatment_col, treatment_labels)
-            )
-
-    # two tailed test, so check where absolute values of the null distribution
-    # are greater or equal to the absolute value of the observed difference
-    pval = np.where((np.array(np.abs(null_distribution)) >= np.abs(truediff)))[
-        0
-    ].size / len(null_distribution)
+    if alternative == "two-sided":
+        pval = p_two
+    elif alternative == "less":
+        pval = p_less
+    elif alternative == "greater":
+        pval = p_greater
 
     if pval == 0:
         pval += 1 / (total)
@@ -828,12 +612,13 @@ def multi_sample_test(
     >>> multi_sample_test(data, treatment_col=0, hypotheses="all",
     ...                   correction=None, bootstraps=1000,
     ...                   permutations="all", random_state=111)
-    array([[2.0, 3.0, 0.0355],
-           [1.0, 3.0, 0.0394],
-           [3.0, 4.0, 0.0407],
-           [2.0, 4.0, 0.1477],
-           [1.0, 2.0, 0.4022],
-           [1.0, 4.0, 0.4559]], dtype=object)
+      Condition 1 Condition 2 p-value
+    0         2.0         3.0  0.0354
+    1         1.0         3.0  0.0393
+    2         3.0         4.0  0.0406
+    3         2.0         4.0  0.1476
+    4         1.0         2.0  0.4021
+    5         1.0         4.0  0.4558
 
     Multiple comparison correction to control False Discovery Rate is advisable in
     this situation. The final column now shows the q-values, or "adjusted" p-values
@@ -842,13 +627,14 @@ def multi_sample_test(
     >>> multi_sample_test(data, treatment_col=0, hypotheses="all",
     ...                   correction='fdr', bootstraps=1000,
     ...                   permutations="all", random_state=111)
-    array([[2.0, 3.0, 0.0355, 0.0814],
-           [1.0, 3.0, 0.0394, 0.0814],
-           [3.0, 4.0, 0.0407, 0.0814],
-           [2.0, 4.0, 0.1477, 0.22155],
-           [1.0, 2.0, 0.4022, 0.4559],
-           [1.0, 4.0, 0.4559, 0.4559]], dtype=object)
-
+      Condition 1 Condition 2 p-value Corrected p-value
+    0         2.0         3.0  0.0354            0.0812
+    1         1.0         3.0  0.0393            0.0812
+    2         3.0         4.0  0.0406            0.0812
+    3         2.0         4.0  0.1476            0.2214
+    4         1.0         2.0  0.4021            0.4558
+    5         1.0         4.0  0.4558            0.4558
+    
     Perhaps the experimenter is not interested in every pairwise comparison - perhaps
     condition 2 is a control that all other conditions are meant to be compared to.
     The comparisons of interest can be specified using a list.
@@ -857,9 +643,10 @@ def multi_sample_test(
     >>> multi_sample_test(data, treatment_col=0, hypotheses=tests,
     ...                   correction='fdr', bootstraps=1000,
     ...                   permutations="all", random_state=222)
-    array([[2.0, 3.0, 0.036, 0.108],
-           [2.0, 4.0, 0.1506, 0.2259],
-           [2.0, 1.0, 0.4036, 0.4036]], dtype=object)
+      Condition 1 Condition 2 p-value Corrected p-value
+    0         2.0         3.0  0.0359            0.1077
+    1         2.0         4.0  0.1505           0.22575
+    2         2.0         1.0  0.4035            0.4035
 
 
     """
@@ -907,7 +694,7 @@ def multi_sample_test(
             (data[:, treatment_col] == output[i, 0]),
             (data[:, treatment_col] == output[i, 1]),
         )
-        output[i, 2] = two_sample_test(
+        output[i, 2] = hypothesis_test(
             data[test_idx],
             treatment_col=treatment_col,
             compare=compare,
@@ -929,6 +716,12 @@ def multi_sample_test(
         out[:, :-1] = output
         out[:, -1] = q_vals
         output = out
+        output = pd.DataFrame(
+            output,
+            columns=["Condition 1", "Condition 2", "p-value", "Corrected p-value"],
+        )
+    else:
+        output = pd.DataFrame(output, columns=["Condition 1", "Condition 2", "p-value"])
 
     return output
 
@@ -1118,31 +911,32 @@ def confidence_interval(
 
     >>> confidence_interval(data, treatment_col=0, interval=95, 
     ...    bootstraps=1000, permutations='all', random_state=1)
-    (1.3142402418925743, 6.125225510899227)
+    (1.314807450602109, 6.124658302189696)
 
     The true difference is 2, which falls within the interval. We can examine
     the p-value for the corresponding dataset:
 
-    >>> from hierarch.stats import linear_regression_test
-    >>> linear_regression_test(data, treatment_col=0,
+    >>> from hierarch.stats import hypothesis_test
+    >>> hypothesis_test(data, treatment_col=0, compare='corr',
     ...                 bootstraps=1000, permutations='all',
     ...                 random_state=1)
-    0.013714285714285714
+    0.013685714285714285
 
-    This suggests that while the 95% confidence interval does not contain 0, the 99.9%
+    This suggests that while the 95% confidence interval does not contain 0, the 99.5%
     confidence interval should.
 
-    >>> confidence_interval(data, treatment_col=0, interval=99.9, 
+    >>> confidence_interval(data, treatment_col=0, interval=99.5, 
     ...    bootstraps=1000, permutations='all', random_state=1)
-    (-2.452406804001683, 9.89187255679354)
+    (-0.12320618535452432, 7.56267193814634)
 
-    hierarch.stats.two_sample_test can be used to generate the null distribution by
-    specifying compare = "means". This should return a very similar interval.
+    A permutation t-test can be used to generate the null distribution by
+    specifying compare = "means". This should return the same or a very
+    similar interval.
 
     >>> confidence_interval(data, treatment_col=0, interval=95, 
     ...    compare='means', bootstraps=1000, 
     ...    permutations='all', random_state=1)
-    (1.330299483004123, 6.109166269787682)
+    (1.314807450602109, 6.124658302189696)
 
     Setting compare = "corr" will generate a confidence interval for the slope
     in a regression equation.     
@@ -1156,25 +950,22 @@ def confidence_interval(
     >>> confidence_interval(data, treatment_col=0, interval=95,
     ...                 compare='corr', bootstraps=100,
     ...                 permutations=1000, random_state=1)
-    (0.7521535040639042, 1.6431732504510854)
+    (0.7712039924329259, 1.5597743222883649)
 
     The dataset was specified to have a true slope of 1, which is within the interval.
 
     """
 
-    tests = {"means": two_sample_test, "corr": linear_regression_test}
+    rng = np.random.default_rng(random_state)
+
+    alpha = (100 - interval) / 200
 
     # turns the input array or dataframe into a float64 array
     if isinstance(data_array, (np.ndarray, pd.DataFrame)):
-        data = preprocess_data(data_array)
+        data = _preprocess_data(data_array)
     else:
         raise TypeError("Input data must be ndarray or DataFrame.")
 
-    try:
-        hypothesis_test = tests[compare]
-    except KeyError:
-        raise KeyError("No such comparison.")
-
     # first compute the null distribution against the null that the effect size is equal to the MLE
     null_imposed_data = data.copy()
     levels_to_agg = data.shape[1] - treatment_col - 3
@@ -1199,49 +990,144 @@ def confidence_interval(
         permutations=permutations,
         kind=kind,
         return_null=True,
-        random_state=random_state,
+        random_state=rng,
     )
 
     target_agg = grouper.transform(data, iterations=levels_to_agg)
 
-    # the null distribution has a tendency to underestimate the "true" null's variance
-    # by recomputing it a few times, we get the correct coverage
-    for i in range(iterations - 1):
+    # make a guess as to the lower and upper bounds of the confidence interval
 
-        null_agg = grouper.transform(null_imposed_data, iterations=levels_to_agg)
+    null_agg = grouper.transform(null_imposed_data, iterations=levels_to_agg)
 
-        lower, upper = _compute_interval(
-            np.array(null), null_agg, target_agg, treatment_col, interval
+    current_lower = _compute_interval(
+        np.array(null), null_agg, target_agg, treatment_col, alpha
+    )
+    current_upper = _compute_interval(
+        np.array(null), null_agg, target_agg, treatment_col, 1 - alpha
+    )
+
+    # refine the bounds via iterative hypothesis testing
+    # each bound needs to be found separately
+
+    # find lower bound
+
+    if compare == "means":
+        alternative_lower, alternative_upper = "less", "greater"
+    else:
+        alternative_lower, alternative_upper = "greater", "less"
+
+    for i in range(iterations):
+
+        bound_imposed_data = data.copy()
+        bound_imposed_data[:, -1] -= (
+            current_lower * bound_imposed_data[:, treatment_col]
         )
-
-        if np.abs(interval - (1 - _) * 100) < tolerance:
-            break
-
-        null_imposed_data = data.copy()
-        null_imposed_data[:, -1] -= lower * null_imposed_data[:, treatment_col]
-        _, null = hypothesis_test(
-            null_imposed_data,
+        current_p, null = hypothesis_test(
+            bound_imposed_data,
             treatment_col,
+            compare=compare,
+            alternative=alternative_lower,
             skip=skip,
-            bootstraps=50,
-            permutations=100,
+            bootstraps=bootstraps,
+            permutations=permutations,
             kind=kind,
             return_null=True,
-            random_state=random_state,
+            random_state=rng,
         )
 
-    _, null = hypothesis_test(
-        null_imposed_data,
-        treatment_col,
-        skip=skip,
-        bootstraps=bootstraps,
-        permutations=permutations,
-        kind=kind,
-        return_null=True,
-        random_state=random_state,
-    )
+        if np.abs(100 * (alpha - current_p)) < tolerance:
+            break
 
-    return _compute_interval(
-        np.array(null), null_agg, target_agg, treatment_col, interval
-    )
+        bound_agg = grouper.transform(bound_imposed_data, iterations=levels_to_agg)
 
+        current_lower = _compute_interval(
+            np.array(null), bound_agg, target_agg, treatment_col, alpha
+        )
+
+    else:
+        warn(
+            " ".join(["lower tail:", str(current_p), "failed to converge"]),
+            ConvergenceWarning,
+            stacklevel=2,
+        )
+
+    for i in range(iterations):
+
+        bound_imposed_data = data.copy()
+        bound_imposed_data[:, -1] -= (
+            current_upper * bound_imposed_data[:, treatment_col]
+        )
+        current_p, null = hypothesis_test(
+            bound_imposed_data,
+            treatment_col,
+            compare=compare,
+            alternative=alternative_upper,
+            skip=skip,
+            bootstraps=bootstraps,
+            permutations=permutations,
+            kind=kind,
+            return_null=True,
+            random_state=rng,
+        )
+
+        if np.abs(100 * (alpha - current_p)) < tolerance:
+            break
+        bound_agg = grouper.transform(bound_imposed_data, iterations=levels_to_agg)
+
+        current_upper = _compute_interval(
+            np.array(null), bound_agg, target_agg, treatment_col, 1 - alpha
+        )
+
+    else:
+        warn(
+            " ".join(["upper tail:", str(current_p), "failed to converge"]),
+            ConvergenceWarning,
+            stacklevel=2,
+        )
+
+    return current_lower, current_upper
+
+
+class ConvergenceWarning(Warning):
+    def __init__(self, message):
+        self.message = message
+
+    def __str__(self):
+        return repr(self.message)
+
+
+@jit(nopython=True, cache=True)
+def _compute_interval(null, null_data, target_data, treatment_col, alpha):
+
+    x = null_data[:, treatment_col]
+    y = null_data[:, -1]
+
+    denom = _cov_std_error(x, y)
+    bound = np.quantile(null, alpha) * denom / bivar_central_moment(x, x)
+
+    x = target_data[:, treatment_col]
+    y = target_data[:, -1]
+    denom = _cov_std_error(x, y)
+
+    bound = bound + (bivar_central_moment(x, y) / bivar_central_moment(x, x))
+    return bound
+
+
+@jit(nopython=True, cache=True)
+def _cov_std_error(x, y):
+    n = len(x)
+    # first term is the second symmetric bivariate central moment. an approximate
+    # bias correction of n - root(2) is applied
+    denom_1 = bivar_central_moment(x, y, pow=2, ddof=2 ** 0.5)
+
+    # second term is the product of the standard deviations of x and y over n - 1.
+    # this term rapidly goes to 0 as n goes to infinity
+    denom_2 = (
+        bivar_central_moment(x, x, pow=1, ddof=1)
+        * bivar_central_moment(y, y, pow=1, ddof=1)
+    ) / (n - 1)
+
+    # third term is the square of the covariance of x and y. an approximate bias
+    # correction of n - root(3) is applied
+    denom_3 = ((n - 2) * (bivar_central_moment(x, y, pow=1, ddof=1.75) ** 2)) / (n - 1)
+    return ((1 / (n - 1.5)) * (denom_1 + denom_2 - denom_3)) ** 0.5
